
local VERSION = GetAddOnMetadata("RandomCompanion", "Version");

RandomCompanion = {
    mounts = {};
    critters = {};
    Initialized = false;
    FlyingTrained = false
};

local function debugmsg(msg, debuglevel)
    if debuglevel == nil then
        debuglevel = 1;
    end
    if tonumber(debuglevel) then
        local debuglevel = tonumber(debuglevel);
    else    
        local debuglevel = 1;
    end
    if RandomCompanion_Settings.DEBUG and (debuglevel >= tonumber(RandomCompanion_Settings.DEBUGLevel)) then
        DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion Debug[" .. debuglevel .. "]: " .. msg);
    end
end

function RandomCompanion.Event(self, event, ...)
    if event == "PLAYER_ENTERING_WORLD" then
        if RandomCompanion_Settings.Quiet and not RandomCompanion.Initialized then
            RandomCompanion.Initialize(false);
        elseif not RandomCompanion.Initialized then
            RandomCompanion.Initialize(true);
        end
    elseif event == "MOUNT_JOURNAL_USABILITY_CHANGED" then
        RC_REQUIRE_LOAD_MOUNTS = true;
    elseif event == "COMPANION_LEARNED" or event == "PET_JOURNAL_LIST_UPDATE" or event == "PET_JOURNAL_PET_DELETED" then
        debugmsg("Reload required due to event: " .. event, 3);
        RC_REQUIRE_RELOAD = true;
    elseif event == "COMBAT_LOG_EVENT_UNFILTERED" and RandomCompanion_Settings.Traincrusher then
        if #RandomCompanion.critters.traincrusher > 0 then
            local timestamp, eventtype, hideCaster, srcGUID, srcName, srcFlags, srcFlags2, dstGUID, dstName, dstFlags, dstFlags2, spellid = select(1, ...);
            if eventtype == "SPELL_CREATE" and spellid == 61031 then
                if not RandomCompanion_Settings.Quiet then
                    DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion: Toy Train Set detected! Created by " .. srcName);
                end
                debugmsg("Toy Train Set detected! Created by " .. srcName, 5);
                if not RandomCompanion.Busy() then
                    local petindex = random(#RandomCompanion.critters.traincrusher);
                    debugmsg("Crushing train with " .. RandomCompanion.critters.traincrusher[petindex].name .. " . petID is " .. RandomCompanion.critters.traincrusher[petindex].petID, 10);
                    RandomCompanion.origC_PetJournal_SummonPetByGUID(RandomCompanion.critters.traincrusher[petindex].petID);
                end
            end
        end
    else
        debugmsg("Unknown event triggered.", 10);
    end
end

function RandomCompanion.Help()
    --For backwards compatibility with old slash commands
    RandomCompanion.Slash("help");
end

function RandomCompanion.Initialize(showstatus)
    RC_REQUIRE_RELOAD = false;
    
    if not RandomCompanion_Settings then
        RandomCompanion_Settings = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()] then
        RandomCompanion_Settings[GetRealmName()] = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()][UnitName("player")] then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")] = {};
    end
    
    if not RandomCompanion_Settings.DEBUG then
        RandomCompanion_Settings.DEBUG = false;
    end
    
    if RandomCompanion_Settings.DEBUGLevel == nil then
        RandomCompanion_Settings.DEBUGLevel = 1;
    end
    
    if RandomCompanion_Settings.AccountWideWeights == nil then
        RandomCompanion.AccountWideWeights = true;
    end
    
    if RandomCompanion_Settings.AutoRecall == nil then
        RandomCompanion_Settings.AutoRecall = false;
        if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RandomChange or RandomCOmpanion_settings.RaidDismiss) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
            RandomCompanion.StartPetTick()
        else
            RandomCompanion.StopPetTick()
        end
    end
    
    if RandomCompanion_Settings.AutoDismiss == nil then
        RandomCompanion_Settings.AutoDismiss = true;
        if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RandomChange or RandomCompanion_Settings.RaidDismiss) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
            RandomCompanion.StartPetTick()
        else
            RandomCompanion.StopPetTick()
        end
    end
    
    if RandomCompanion_Settings.AllFlyingOnGround == nil then
        RandomCompanion_Settings.AllFlyingOnGround = false;
    end
    
    if RandomCompanion_Settings.RaidDismiss == nil then
        RandomCompanion_Settings.RaidDismiss = false;
    end
    
    if RandomCompanion_Settings.RandomRecall == nil then
        RandomCompanion_Settings.RandomRecall = true;
    end
    
    if RandomCompanion_Settings.RandomChange == nil then
        RandomCompanion_Settings.RandomChange = true;
    end
    
    if RandomCompanion_Settings.RandomChangeTime == nil then
        RandomCompanion_Settings.RandomChangeTime = 15;
    end
    
    if RandomCompanion_Settings.Cloning == nil then
        RandomCompanion_Settings.Cloning = true;
    end
    
    if RandomCompanion_Settings.Traincrusher == nil then
        RandomCompanion_Settings.Traincrusher = false;
    end
    
    if not RandomCompanion_Settings.accountwide then
        RandomCompanion_Settings.accountwide = {};
    end
    
    if not RandomCompanion_Settings.accountwide.weights then
        RandomCompanion_Settings.accountwide.weights = {};
    end

    if not RandomCompanion_Settings.accountwide.FlyingOnGroundMounts then
        RandomCompanion_Settings.accountwide.FlyingOnGroundMounts = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()] then
        RandomCompanion_Settings[GetRealmName()] = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()][UnitName("player")] then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")] = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights = {};
    end

    if not RandomCompanion_Settings[GetRealmName()][UnitName("player")].FlyingOnGroundMounts then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")].FlyingOnGroundMounts = {};
    end
    
    if RandomCompanion_Settings.Quiet == nil then
        RandomCompanion_Settings.Quiet = false;
    end

    RandomCompanion.LoadMounts();
    
    RandomCompanion.critters = {
        weighted = {};
        unweighted = {};
        winter = {};
        byname = {};
        traincrusher = {};
        favorites = {};
    };
    
    if not RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter = nil;
    end
    
    local scrollFrame = PetJournal.listScroll;
    local offset = HybridScrollFrame_GetOffset(scrollFrame);
    local petButtons = scrollFrame.buttons;
    local isWild = PetJournal.isWild;
    local numPets, numOwned = C_PetJournal.GetNumPets(isWild);
    local summonedPetID = C_PetJournal.GetSummonedPetGUID();
    
    for i = 1, C_PetJournal.GetNumPets(PetJournal.isWild) do
        index = offset + i;
        if index <= numPets then
            local petID, speciesID, isOwned, customName, level, favorite, isRevoked, name, icon, petType, creatureID, sourceText, description, isWildPet, canBattle = C_PetJournal.GetPetInfoByIndex(index, isWild);
            local companionID = "pet." .. name .. "." .. speciesID .. "." .. (customName and customName or name)
            
            if not name then
                --This seems to happen when the companion is not in the cache
                RC_REQUIRE_RELOAD = true;
                debugmsg(petID .. " petID critter has no name?", 1);
            elseif isOwned then
                if RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[name] ~= nil then --Import from pre-5.0 version of RandomCompanion
                    RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companionID] = RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[name];
                    RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[name] = nil;
                end
                
                if not RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companionID] then --Check to import early v6.0 weight
                    local vSixOhID = "pet." .. name .. "." .. "nil";
                    if RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[vSixOhID] then
                        RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companionID] = RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[vSixOhID];
                        RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[vSixOhID] = nil;
                    end
                end
                
                if not RandomCompanion_Settings.accountwide.weights[companionID] then --Check to import early v6.0 weight
                    local vSixOhID = "pet." .. name .. "." .. "nil";
                    if RandomCompanion_Settings.accountwide.weights[vSixOhID] then
                        RandomCompanion_Settings.accountwide.weights[companionID] = RandomCompanion_Settings.accountwide.weights[vSixOhID];
                        RandomCompanion_Settings.accountwide.weights[vSixOhID] = nil;
                    end
                end
                
                if not RandomCompanion_Settings.accountwide.weights[companionID] then --import from petID weights
                    local petGUID = "pet." .. name .. "." .. tostring(petID, 10);
                    if RandomCompanion_Settings.accountwide.weights[petGUID] then
                        RandomCompanion_Settings.accountwide.weights[companionID] = RandomCompanion_Settings.accountwide.weights[petGUID];
                        RandomCompanion_Settings.accountwide.weights[petGUID] = nil;
                    end
                end
                
                
                if not RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companionID] then --Set default weight
                    if creatureID == 49590 or creatureID == 49588 or creatureID == 49587 or creatureID == 49586 then
                        --Guild page/herald should be disabled by default
                        RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companionID] = 0;
                    end
                end
                
                if not RandomCompanion_Settings.accountwide.weights[companionID] then --Set accountwide default weight
                    if creatureID == 49590 or creatureID == 49588 or creatureID == 49587 or creatureID == 49586 then
                        --Guild page/herald should be disabled by default
                        RandomCompanion_Settings.accountwide.weights[companionID] = 0;
                    end
                end
                
                -- Don't add same pet twice, corner case is 3 pets of same species are added in sequence of custome name A, B, A, where 2 A's are same level
                if not RandomCompanion.critters.byname[name] or (RandomCompanion.critters.byname[name].customName ~= customName) then
                    local weight = RandomCompanion.GetWeight(companionID);
                    
                    RandomCompanion.critters.byname[name] = {
                        creatureid = creatureID;
                        name = name;
                        index = index;
                        icon = icon;
                        active = active;
                        petID = petID;
                        customName = customName;
                        favorite = favorite;
                        level = level;
                        speciesID = speciesID;
                    }
                    
                    local nextindex = #RandomCompanion.critters.unweighted + 1;
                    RandomCompanion.critters.unweighted[nextindex] = {
                        creatureid = creatureID;
                        name = name;
                        index = index;
                        icon = icon;
                        active = active;
                        petID = petID;
                        customName = customName;
                        favorite = favorite;
                        level = level;
                        speciesID = speciesID;
                    }
                    
                    if favorite then
                        local nextindex = #RandomCompanion.critters.favorites + 1;
                        RandomCompanion.critters.favorites[nextindex] = {
                            creatureid = creatureID;
                            name = name;
                            index = index;
                            icon = icon;
                            active = active;
                            petID = petID;
                            customName = customName;
                            favorite = favorite;
                            level = level;
                            speciesID = speciesID;
                        }
                    end
                    
                    if creatureID == 24968 or creatureID == 50468 or creatureID == 50384 then
                        local nextindex = #RandomCompanion.critters.traincrusher + 1;
                        RandomCompanion.critters.traincrusher[nextindex] = {
                            creatureid = creatureID;
                            name = name;
                            index = index;
                            spellid = spellid;
                            icon = icon;
                            active = active;
                            petID = petID;
                            customName = customName;
                            favorite = favorite;
                            level = level;
                            speciesID = speciesID;
                        }
                    end

                    if weight ~= 1 then
                        debugmsg("Critter " .. name .. " has non-default weight: " .. weight, 1);
                    end
                    
                    if weight > 0 then
                        for counter = 1, weight do
                            local nextindex = #RandomCompanion.critters.weighted + 1;
                            RandomCompanion.critters.weighted[nextindex] = {
                                creatureid = creatureid;
                                name = name;
                                index = index;
                                spellid = spellid;
                                icon = icon;
                                active = active;
                                petID = petID;
                                customName = customName;
                                favorite = favorite;
                                level = level;
                                speciesID = speciesID;
                            }
                        end
                    end
                end
            end
        end
    end
    
    if RandomCompanion_Settings.Traincrusher and #RandomCompanion.critters.traincrusher then
        RandomCompanion.EventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
    else
        RandomCompanion.EventFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
    end
    
    RandomCompanion.SetActiveCompanions();
    RandomCompanion.Initialized = true;
    
    if showstatus then
        RandomCompanion.Status();
    end
end

function RandomCompanion.LoadMounts()
    RC_REQUIRE_LOAD_MOUNTS = false;

    RandomCompanion.mounts = {
        ground = {};
        flying = {};
        flyingonground = {};
        passengerground = {};
        passengerflying = {};
        passengerflyingonground = {};
        aquatic = {};
        aquaticslow = {};
        waterwalking = {};
        vashjir = {};
        qiraji = {};
        lowlevel = {};

        all = {};
    };

    RandomCompanion.FlyingTrained = IsSpellKnown(RC_SPELLID_EXPERT_RIDING) or IsSpellKnown(RC_SPELLID_ARTISAN_RIDING) or IsSpellKnown(RC_SPELLID_MASTER_RIDING);
    RandomCompanion.BfAPathfinderCompleted = IsSpellKnown(RC_SPELLID_BFA_PATHFINDER);

    for index,mountID in pairs(C_MountJournal.GetMountIDs()) do
        local name, spellid, _, _, isUsable, _, _, _, _, _, isCollected = C_MountJournal.GetMountInfoByID(mountID);
        
        if name and isUsable and isCollected then
            
            local companionID = "mount." .. name .. "." .. spellid;
            
            if RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[name] ~= nil then --Import from pre-5.0 version of RandomCompanion
                RandomCompanion.SetWeight(companionID, RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[name]);
                RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[name] = nil;
            end
            
            local weight = RandomCompanion.GetWeight(companionID);
            
            local mountdata = {
                name = name;
                spellid = spellid;
                mountID = mountID;
                weight = weight;
            }
            
            RandomCompanion.mounts.all[spellid] = mountdata;

            --[[https://wow.gamepedia.com/API_C_MountJournal.GetMountInfoExtraByID
                230 for most ground mounts
                231 for  [Riding Turtle] and  [Sea Turtle]
                232 for  [Vashj'ir Seahorse] (was named Abyssal Seahorse prior to Warlords of Draenor)
                241 for Blue, Green, Red, and Yellow Qiraji Battle Tank (restricted to use inside Temple of Ahn'Qiraj)
                242 for Swift Spectral Gryphon (hidden in the mount journal, used while dead in certain zones)
                247 for Red Flying Cloud
                248 for most flying mounts, including those that change capability based on riding skill
                254 for Subdued Seahorse,  [Brinedeep Bottom-Feeder] and  [Fathom Dweller]
                269 for Azure and Crimson Water Strider
                284 for Chauffeured Mekgineer's Chopper and Chauffeured Mechano-Hog
            ]]

            local _, _, _, _, mountType = C_MountJournal.GetMountInfoExtraByID(mountID);
            local flyingOnGround = RandomCompanion.GetFlyingOnGround(companionID);
            local passenger = RC_MOUNT_IDS.passenger[spellid];

            -- flying mounts
            if mountType == RC_MOUNTTYPE_FLYING or mountType == RC_MOUNTTYPE_FLYINGCLOUD then
                
                -- flying passenger mounts
                if passenger then
                    if (flyingOnGround or RandomCompanion_Settings.AllFlyingOnGround) then
                        table.insert(RandomCompanion.mounts.passengerflyingonground, mountdata);
                    else
                        table.insert(RandomCompanion.mounts.passengerflying, mountdata);
                    end
                -- flying mounts
                else
                    if (flyingOnGround or RandomCompanion_Settings.AllFlyingOnGround) then
                        table.insert(RandomCompanion.mounts.flyingonground, mountdata);
                    else
                        table.insert(RandomCompanion.mounts.flying, mountdata);
                    end
                end

            -- slow aquatic (sea turtle)
            elseif RC_MOUNT_IDS.aquaticslow[spellid] then
                table.insert(RandomCompanion.mounts.aquaticslow, mountdata);

            -- aquatic mounts
            elseif mountType == RC_MOUNTTYPE_AQUATIC then
                table.insert(RandomCompanion.mounts.aquatic, mountdata);

            -- waterwalking mounts
            elseif mountType == RC_MOUNTTYPE_WATERWALKING then
                table.insert(RandomCompanion.mounts.waterwalking, mountdata);

            -- Vashj'ir Seahorse
            elseif mountType == RC_MOUNTTYPE_VASHJIR then
                table.insert(RandomCompanion.mounts.vashjir, mountdata);

            -- Qiraji Battle Tanks
            elseif mountType == RC_MOUNTTYPE_QIRAJI then
                table.insert(RandomCompanion.mounts.qiraji, mountdata);

            -- Chauffeured heirloom mounts
            elseif mountType == RC_MOUNTTYPE_LOWLEVEL then
                table.insert(RandomCompanion.mounts.lowlevel, mountdata);

            -- default: ground mounts
            else -- if mountType == RC_MOUNTTYPE_GROUND then
                -- passengerground
                if passenger then
                    table.insert(RandomCompanion.mounts.passengerground, mountdata);
                -- ground
                else
                    table.insert(RandomCompanion.mounts.ground, mountdata);
                end
            end
        end
    end
end

function RandomCompanion.Debug()
    --For backwards compatibility with old slash commands
    RandomCompanion.Slash("debug");
end

function RandomCompanion.Status()
    local loadmessage = "";

    local mountCount = 0;
    for _ in pairs(RandomCompanion.mounts.all) do mountCount = mountCount + 1 end;
    
    if mountCount > 0 then
        if loadmessage ~= "" then
            loadmessage = loadmessage .. ", ";
        end
        loadmessage = loadmessage .. mountCount .. " mounts";
    end

    if #RandomCompanion.critters.unweighted > 0 then
        if loadmessage ~= "" then
            loadmessage = loadmessage .. ", ";
        end
        loadmessage = loadmessage .. #RandomCompanion.critters.unweighted .. " critters";
    end
    
    if loadmessage ~= "" then
        loadmessage = " with " .. loadmessage .. ".";
    end
    
    if RandomCompanion_Settings.RandomRecall then
        loadmessage = loadmessage .. " Vanity pets will be automatically randomly re-summoned when possible.";
    elseif RandomCompanion_Settings.AutoRecall then
        loadmessage = loadmessage .. " Vanity pets will be automatically re-summoned when possible.";
    end
    
    if RandomCompanion_Settings.RandomChange then
        loadmessage = loadmessage .. " Vanity pets will be randomly changed every " .. RandomCompanion_Settings.RandomChangeTime .. " minutes.";
        if RandomCompanion.LastCritterChange == nil then
            RandomCompanion.LastCritterChange = GetTime();
        end
        local nextchange = RandomCompanion.LastCritterChange - (GetTime() - (RandomCompanion_Settings.RandomChangeTime * 60));
        if nextchange < 0 then
            loadmessage = loadmessage .. " Next change will occur as soon as you are not busy.";
        else
            loadmessage = loadmessage .. " Next change will occur in " .. floor(nextchange / 60) .. ":" .. (floor(nextchange % 60)) .. ".";
        end
    end
    
    if RandomCompanion_Settings.Cloning then
        loadmessage = loadmessage .. " Pets and mounts will be cloned when possible.";
    end
    
    if RandomCompanion_Settings.Traincrusher and #RandomCompanion.critters.traincrusher > 0 then
        loadmessage = loadmessage .. " Trains will be crushed!";
    end
    
    if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
        RandomCompanion.StartPetTick()
    else
        RandomCompanion.StopPetTick()
    end
    
    if RandomCompanion_Settings.AutoDismiss then
        loadmessage = loadmessage .. " Vanity pets will be automatically dismissed when entering stealth while flagged for PVP.";
        if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
            RandomCompanion.StartPetTick()
        else
            RandomCompanion.StopPetTick()
        end
    end
    
    if RandomCompanion_Settings.RaidDismiss then
        loadmessage = loadmessage .. " Vanity pets will be automatically dismissed in raids.";
        if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
            RandomCompanion.StartPetTick()
        else
            RandomCompanion.StopPetTick()
        end
    end
    
    if RC_REQUIRE_RELOAD then
        loadmessage = loadmessage .. " RandomCompanion will reload when next companion is summoned.";
    end
    
    DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion " .. VERSION .. " loaded" .. loadmessage);
end

function RandomCompanion.InsertToWeightedTable(weighttable, weight, mountdata)
    if weight > 0 then
        for i = 1, weight do
            table.insert(weighttable, mountdata);
        end
    end
end

function AddToWeightedTable(t1, t2)
    for k,v in pairs(t2) do
        if v.weight and v.mountID then
            for i = 1, v.weight do
                table.insert(t1, v.mountID);
            end
        end
    end 
 
   return t1
end

function RandomCompanion.Mount(cmd)
    local targetUsableMountID;
    local playerInCombat = InCombatLockdown();
    local mountlist = {};

    if (cmd) then
        --Parse any macro conditionals such as [modifier:ctrl] or [combat]
        cmd = SecureCmdOptionParse(cmd);
    end

    -- no usable mount in combat, no use loading
    if not playerInCombat then
        if next(RandomCompanion.mounts.all) == nil  then
            RC_REQUIRE_LOAD_MOUNTS = true;
        end

        if RC_REQUIRE_LOAD_MOUNTS then
            RandomCompanion.LoadMounts();
        end

        if not cmd and RandomCompanion_Settings.Cloning then
            targetUsableMountID = RandomCompanion.GetTargetUsableMountID();
        end
    end
    
    if IsMounted() then
        Dismount();
    elseif CanExitVehicle() then
        VehicleExit();
    elseif playerInCombat then
        -- TODO: Druid forms?
        UIErrorsFrame:AddMessage("You are in combat.", 1.0, 0.0, 0.0, 53, 5);
    elseif targetUsableMountID then
        C_MountJournal.SummonByID(targetUsableMountID);
    else
        local flyable = RandomCompanion.IsFlyable();
        mountlist = {};

        -- Passenger mounts
        if cmd == "passenger" then
            AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerflyingonground);
            if flyable then
                AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerflying);
            end
            -- ground passenger mounts when in no fly area or in fly area but no flying passenger mounts
            if not flyable or #mountlist == 0 then
                AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerground);
            end

        -- Flying mounts
        elseif cmd == "flying" then
            AddToWeightedTable(mountlist, RandomCompanion.mounts.flying);
            AddToWeightedTable(mountlist, RandomCompanion.mounts.flyingonground);
            AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerflying);
            AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerflyingonground);

        -- Ground mounts
        elseif cmd == "ground" then
            AddToWeightedTable(mountlist, RandomCompanion.mounts.ground);
            AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerground);

        -- Waterwalking mounts
        elseif cmd == "waterwalking" then
            AddToWeightedTable(mountlist, RandomCompanion.mounts.waterwalking);

        -- Aquatic mounts
        elseif IsSwimming() or IsSubmerged() or cmd == "aquatic" then
            if RandomCompanion.PlayerUnderMapID(RC_MAPID_VASHJIR) then
                AddToWeightedTable(mountlist, RandomCompanion.mounts.vashjir);
            end
            
            AddToWeightedTable(mountlist, RandomCompanion.mounts.aquatic);
            
            -- use slow aquatic mount only if no other available
            if #mountlist == 0 then
                AddToWeightedTable(mountlist, RandomCompanion.mounts.aquaticslow);
            end

        -- Qiraji Battle Tanks
        elseif #RandomCompanion.mounts.qiraji>0 then
            AddToWeightedTable(mountlist, RandomCompanion.mounts.qiraji);

        end

        -- Default mount command outside of water
        if #mountlist==0 then
            -- Flying mounts
            if flyable then
                AddToWeightedTable(mountlist, RandomCompanion.mounts.flying);
                AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerflying);  
            end

            -- Flying on ground mounts used everywhere
            AddToWeightedTable(mountlist, RandomCompanion.mounts.flyingonground);
            AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerflyingonground);

            -- In Nazjatar with Budding Deepcoral
            if C_Map.GetBestMapForUnit("player") == RC_MAPID_NAZJATAR and C_QuestLog.IsQuestFlaggedCompleted(RC_QUESTID_BUDDING_DEEPCORAL) then
                AddToWeightedTable(mountlist, RandomCompanion.mounts.aquatic);
            end

            -- Ground mounts when in no fly zone or no other mounts available
            if not flyable or #mountlist == 0 then
                AddToWeightedTable(mountlist, RandomCompanion.mounts.ground);
                AddToWeightedTable(mountlist, RandomCompanion.mounts.passengerground);
            end
        end

        -- No mounts usable, try low level mounts (Chauffeured)
        if #mountlist == 0 and #RandomCompanion.mounts.lowlevel > 0 then
            AddToWeightedTable(mountlist, RandomCompanion.mounts.lowlevel);
        end
        
        if #mountlist > 0 then
            C_MountJournal.SummonByID(mountlist[random(#mountlist)]);
        else
            UIErrorsFrame:AddMessage("No usable mounts.", 1.0, 0.0, 0.0, 53, 5);
        end
    end
end

local __continentID = {}

function RandomCompanion.IsFlyable()
    if not RandomCompanion.FlyingTrained then
        return false;
    end

    if GetRealZoneText() == RC_STRING_WINTERGRASP then
        for areaid = 1, GetNumWorldPVPAreas() do
            local pvpID, localizedName, isActive, canQueue, startTime, canEnter = GetWorldPVPAreaInfo(areaid);
            if localizedName == RC_STRING_WINTERGRASP then
                if isActive then
                    wg_inprogress = 1;
                else
                    wg_inprogress = 0;
                end
                break;
            end
        end
        
        debugmsg("Currently in Wintergrasp. In progress: " .. tonumber(wg_inprogress), 2);
        
        if IsFlyableArea() and wg_inprogress == 0 then
            return true;
        else
            return false;
        end
    end

    local mapID = C_Map.GetBestMapForUnit("player")
    if __continentID[mapID] == nil then
        local continentInfo = MapUtil.GetMapParentInfo(mapID, Enum.UIMapType.Continent, true)

        if continentInfo then 
            __continentID[mapID] = continentInfo.mapID;
        end
    end

    if IsFlyableArea() or __continentID[mapID] == RC_MAPID_DRAENOR or __continentID[mapID] == RC_MAPID_BROKEN_ISLES then
        -- No fly instances reporting can fly
        if RC_NOFLY_INSTANCE_IDS[(select(8,GetInstanceInfo()))] then
            return false;
        end

        -- Warfronts
        if C_Scenario.IsInScenario() then
            local scenarioType = select(10, C_Scenario.GetInfo())
            if scenarioType == LE_SCENARIO_TYPE_WARFRONT then
                return false
            end
        end 

        if __continentID[mapID] == RC_MAPID_ZANDALAR or __continentID[mapID] == RC_MAPID_KUL_TIRAS or mapID == RC_MAPID_NAZJATAR then
            return RandomCompanion.BfAPathfinderCompleted;
        end

        return true;
    else
        return false;
    end
end

function RandomCompanion.PlayerUnderMapID(mapID)
    local mapInfo = C_Map.GetMapInfo(C_Map.GetBestMapForUnit("player"));

    while mapInfo do
        if mapInfo.mapID == mapID then
            return true;
        end
        mapInfo = C_Map.GetMapInfo(mapInfo.parentMapID);
    end

    return false;
end

function RandomCompanion.Pet(cmd)
    local clonetarget;
    
    if cmd and string.lower(cmd) == "autorecall" then
        --For backwards compatibility with old slash commands
        RandomCompanion.Slash(cmd);
    else
        -- Uncomment the next few lines to force clearing the PetJournal filters before summoning pets. This will cause RandomCompanion to reload, and may cause a brief pause while this happens.
        C_PetJournal.SetFilterChecked(LE_PET_JOURNAL_FILTER_COLLECTED, true);
        C_PetJournal.SetFilterChecked(LE_PET_JOURNAL_FILTER_NOT_COLLECTED, false);
        C_PetJournal.SetAllPetTypesChecked(true);
        C_PetJournal.SetAllPetSourcesChecked(true);
        C_PetJournal.ClearSearchFilter();
        
        if RC_REQUIRE_RELOAD then
            --Re-initialize if one of the pet or mount names was not available in the initial load
            debugmsg("Re-initializing RandomCompanion in RandomCompanion.Pet()", 2);
            RandomCompanion.Initialize();
        end
        
        local targetType = UnitCreatureType("target");
        debugmsg("Target is " .. (targetType or "<nil>"), 2);
        
        if RandomCompanion_Settings.Cloning and targetType and (targetType == "Non-combat Pet" or targetType == "Critter") then
            if not RandomCompanionScanningTooltip then
                CreateFrame("GameTooltip", "RandomCompanionScanningTooltip", UIParent, "GameTooltipTemplate");
            end
            RandomCompanionScanningTooltip:AddFontStrings(
                RandomCompanionScanningTooltip:CreateFontString( "$parentTextLeft1", nil, "GameTooltipText" ),
                RandomCompanionScanningTooltip:CreateFontString( "$parentTextRight1", nil, "GameTooltipText" )
            );
            GameTooltip_SetDefaultAnchor(RandomCompanionScanningTooltip, UIParent);
            RandomCompanionScanningTooltip:SetUnit("target");
            RandomCompanionScanningTooltip:Show();
            local targettitle = RandomCompanionScanningTooltipTextLeft2:GetText();
            RandomCompanionScanningTooltip:Hide();
            
            if RandomCompanion.critters.byname[targettitle] then
                --Title should be used first because it's possible that someone could rename their pet to have the same name as another pet
                debugmsg("Found cloneable pet by title " .. targettitle, 5);
                clonetarget = RandomCompanion.critters.byname[targettitle].petID;
            elseif RandomCompanion.critters.byname[targetname] then
                clonetarget = RandomCompanion.critters.byname[targetname].petID;
                debugmsg("Found cloneable pet by name " .. targetname .. ". petID is " .. clonetarget, 5);
            end
        end
        
        --Check if your target is a pet that you have, and clone it
        if clonetarget ~= nil and RandomCompanion_Settings.Cloning then
            debugmsg("Cloning pet: " .. clonetarget .. ".", 10);
            if C_PetJournal.GetSummonedPetGUID() and C_PetJournal.GetSummonedPetGUID() == clonetarget then
                -- if target is the same, dismiss first
                RandomCompanion.TempDismissCompanion("CRITTER")
            end
            C_PetJournal.SummonPetByGUID(clonetarget);           
        elseif #RandomCompanion.critters.weighted > 0 then
            local petindex = nil;
            
            petindex = random(#RandomCompanion.critters.weighted);
            
            if petindex ~= nil then
                if C_PetJournal.GetSummonedPetGUID() and C_PetJournal.GetSummonedPetGUID() == RandomCompanion.critters.weighted[petindex].petID then
                    -- if same pet, dismiss first so player know something happened
                    RandomCompanion.TempDismissCompanion("CRITTER")
                end
                debugmsg("Summoning " .. RandomCompanion.critters.weighted[petindex].name .. ". petID is " .. RandomCompanion.critters.weighted[petindex].petID, 10);
                C_PetJournal.SummonPetByGUID(RandomCompanion.critters.weighted[petindex].petID);
            end
        elseif GetNumCompanions("CRITTER") > 0 then
            --We have pets, but they aren't in the pet list. The most likely issue is that no pets were in the cache when RandomCompanion.Initialize() ran, so let's run it again.
            if not RC_REINITIALIZED then
                debugmsg("Re-initializing because pet names weren't cached yet.");
                RC_REINITIALIZED = true; --We don't want an infinite loop, do we?
                RandomCompanion.Initialize();
                RandomCompanion.Pet();
            else
                RC_REINITIALIZED = nil;
            end
        else
            UIErrorsFrame:AddMessage("You have no vanity pets", 1.0, 0.0, 0.0, 53, 5);
        end
    end
end

function RandomCompanion.SetWeight(companion, value)
    if not RandomCompanion_Settings then
        RandomCompanion_Settings = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()] then
        RandomCompanion_Settings[GetRealmName()] = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()][UnitName("player")] then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")] = {};
    end
    
    if not RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights = {};
    end
    
    if not RandomCompanion_Settings.accountwide then
        RandomCompanion_Settings.accountwide = {};
    end
    
    if not RandomCompanion_Settings.accountwide.weights then
        RandomCompanion_Settings.accountwide.weights = {};
    end
    
    if RandomCompanion_Settings.AccountWideWeights then
        RandomCompanion_Settings.accountwide.weights[companion] = value;
        debugmsg("RandomCompanion_Settings.accountwide.weights["..companion.."] = "..RandomCompanion_Settings.accountwide.weights[companion], 1);
    else
        RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companion] = value;
        debugmsg("RandomCompanion_Settings["..GetRealmName().."]["..UnitName("player").."].weights["..companion.."] = "..RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companion], 1);
    end
end

function RandomCompanion.GetWeight(companion)
    if RandomCompanion_Settings.AccountWideWeights then
        if RandomCompanion_Settings.accountwide.weights[companion] then
            return RandomCompanion_Settings.accountwide.weights[companion];
        else
            return 1;
        end
    else
        if RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companion] then
            return RandomCompanion_Settings[GetRealmName()][UnitName("player")].weights[companion];
        else
            return 1;
        end
    end
end

function RandomCompanion.SetFlyingOnGround(companion, value)
    if RandomCompanion_Settings.AccountWideWeights then
        RandomCompanion_Settings.accountwide.FlyingOnGroundMounts[companion] = value;
    else
        RandomCompanion_Settings[GetRealmName()][UnitName("player")].FlyingOnGroundMounts[companion] = value;
    end
end

function RandomCompanion.GetFlyingOnGround(companion)
    if RandomCompanion_Settings.AccountWideWeights then
        if RandomCompanion_Settings.accountwide.FlyingOnGroundMounts[companion] then
            return RandomCompanion_Settings.accountwide.FlyingOnGroundMounts[companion];
        else
            return RC_MOUNT_IDS.flyingOnGround[tonumber(string.match(companion, "%.([0-9]+)$"))];
        end
    else
        if RandomCompanion_Settings[GetRealmName()][UnitName("player")].FlyingOnGroundMounts[companion] then
            return RandomCompanion_Settings[GetRealmName()][UnitName("player")].FlyingOnGroundMounts[companion];
        else
            return RC_MOUNT_IDS.flyingOnGround[tonumber(string.match(companion, "%.([0-9]+)$"))];
        end
    end
end

function RandomCompanion.SetActiveCompanions(petID)
    local activecritter;

    if petID then
        activecritter = petID;
    else
        activecritter = C_PetJournal.GetSummonedPetGUID();
    end

    if activecritter then
        RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter = activecritter;
        debugmsg("Setting active critter to petID " .. activecritter, 5);
        RandomCompanion.LastCritterChange = GetTime();
    end

    if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
        RandomCompanion.StartPetTick()
    else
        RandomCompanion.StopPetTick()
    end
end

function RandomCompanion.Busy()
    local index = 1;
    local currentBuffName = UnitBuff("player", index);
    local hasBusyBuff = false;
    
    --Check to see if you're doing something that will break when recalling a critter
    while currentBuffName do
        if  RC_BUSY_AURA[currentBuffName] then
            hasBusyBuff = true;
            break;
        end
        index = index + 1;
        currentBuffName = UnitBuff("player", index);
    end
    
    local inInstance, instanceType = IsInInstance();
    
    return InCombatLockdown() or UnitIsDeadOrGhost("player") or UnitCastingInfo("player") or UnitChannelInfo("player") or SpellIsTargeting() or IsMounted() or IsFalling() or IsStealthed() or GetNumLootItems()~= 0 or not HasFullControl() or UnitInVehicle("player") or 
        (instanceType == "raid" and RandomCompanion_Settings.RaidDismiss) or UnitIsAFK("player") or  C_PetBattles.IsInBattle() or (BarberShopFrame and not BarberShopFrame:IsVisible());
end

function RandomCompanion.StartPetTick()
	if not RandomCompanion.PetTickHandle then
		RandomCompanion.PetTickHandle = C_Timer.NewTicker(2, RandomCompanion.PetTick)
	end
end

function RandomCompanion.StopPetTick()
	if RandomCompanion.PetTickHandle then
		RandomCompanion.PetTickHandle:Cancel()
		RandomCompanion.PetTickHandle = nil
	end
end


function RandomCompanion.PetTick()
    if not ((RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter) then
		RandomCompanion.StopPetTick()
    end
    
    if (RandomCompanion.RecallTime == nil or GetTime() > RandomCompanion.RecallTime) and (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
        debugmsg("Checking to recall pets", 1);
        local active = C_PetJournal.GetSummonedPetGUID();
        
        if RandomCompanion.LastCritterChange == nil then
            RandomCompanion.LastCritterChange = GetTime();
        end
        
        if not active then
            if not RandomCompanion.Busy() then
                if RandomCompanion.NotBusyTime == nil then
                    --Start the timer to recall
                    debugmsg("Recalling critter in 5.5 seconds", 2);
                    RandomCompanion.NotBusyTime = GetTime();
                    RandomCompanion.RecallTime = GetTime() + .1;
                end
                if GetTime() > (RandomCompanion.NotBusyTime + 5.5) then 
                    --You haven't been busy for 5.5 seconds, so it should be safe to recall your critter
                    RandomCompanion.NotBusyTime = nil;
                    RandomCompanion.RecallTime = GetTime() + 2;
                    debugmsg("Recalling critter now", 4);
                    if RandomCompanion_Settings.RandomRecall or (RandomCompanion_Settings.RandomChange and RandomCompanion.LastCritterChange < (GetTime() - (RandomCompanion_Settings.RandomChangeTime * 60))) then
                        RandomCompanion.Pet();
                    else
                        -- Choose a new pet if the current one is a pre-6.0 pet GUID
                        if (string.sub(RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter, 0, 2) == "0x" or RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter == tonumber(RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter)) then
                            debugmsg("Pre-6.0 pet GUID found for active critter. Summoning a new critter instead.", 10);
                            RandomCompanion.Pet();
                        else
                            RandomCompanion.origC_PetJournal_SummonPetByGUID(RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter);
                        end
                        RandomCompanion.LastChangeTime = GetTime();
                    end
                end
            else
                RandomCompanion.NotBusyTime = nil;
                RandomCompanion.RecallTime = GetTime() + 2;
                debugmsg("Not recalling critter because you are busy.", 2);
            end
        elseif RandomCompanion_Settings.RandomChange and RandomCompanion.LastCritterChange < (GetTime() - (RandomCompanion_Settings.RandomChangeTime * 60)) then
            if not RandomCompanion.Busy() then
                if RandomCompanion.NotBusyTime == nil then
                    --Start the timer to recall
                    debugmsg("Changing critter in 5.5 seconds", 2);
                    RandomCompanion.NotBusyTime = GetTime();
                    RandomCompanion.RecallTime = GetTime() + .1;
                end
                if GetTime() > (RandomCompanion.NotBusyTime + 5.5) then 
                    --You haven't been busy for 5.5 seconds, so it should be safe to recall your critter
                    RandomCompanion.NotBusyTime = nil;
                    RandomCompanion.RecallTime = GetTime() + 2;
                    debugmsg("Changing critter now", 4);
                    RandomCompanion.Pet();
                end
            else
                RandomCompanion.NotBusyTime = nil;
                RandomCompanion.RecallTime = GetTime() + 2;
                debugmsg("Not changing critter because you are busy.", 1);
            end
        end
    end
    
    if RandomCompanion_Settings.AutoDismiss and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
        debugmsg("Checking to dismiss pets while hiding", 1);
        local stealth, prowl, shadowmeld, invisibility, camouflage, feign;
        local active = C_PetJournal.GetSummonedPetGUID();
        if active then
            local index = 1;
            local currentBuffName = UnitBuff("player", index);
            local isHidden = false;
            
            --Check to see if you're stealthed, prowling, or shadowmelded
            while currentBuffName do
                if  currentBuffName == RC_STRING_STEALTH or
                    currentBuffName == RC_STRING_PROWL or
                    currentBuffName == RC_STRING_SHADOWMELD or
                    currentBuffName == RC_STRING_INVISIBILITY or
                    currentBuffName == RC_STRING_CAMOUFLAGE or
                    currentBuffName == RC_STRING_FEIGN_DEATH then
                    isHidden = true;
                    break;
                end
                index = index + 1;
                currentBuffName = UnitBuff("player", index);
            end
            
            if isHidden and UnitIsPVP("player") then
                debugmsg("Auto-dismissing critter now because you are hiding", 2);
                local activecritter = RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter;
                RandomCompanion.TempDismissCompanion("CRITTER");
            end
        end
    end
    
    if RandomCompanion_Settings.RaidDismiss and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
        debugmsg("Checking to dismiss pets in raid", 1);
        local inInstance, instanceType;
        inInstance, instanceType = IsInInstance();
        local active = C_PetJournal.GetSummonedPetGUID();
        if active and inInstance and instanceType == "raid" then
            debugmsg("Auto-dismissing critter now because you are in a raid", 2);
            local activecritter = RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter;
            RandomCompanion.TempDismissCompanion("CRITTER");
        end
    end
end

function RandomCompanion.GetTargetUsableMountID()
    if UnitIsPlayer("target") then

        local possibleMounts = {}
        local i = 1
        while true do
            local spellID, canApply = select(10, UnitBuff("target", i))
            if spellID and canApply then
                if RandomCompanion.mounts.all[spellID] then
                    return RandomCompanion.mounts.all[spellID].mountID;
                end
            else
                break;
            end
            i = i + 1;
        end
    end
end

function RandomCompanion.PetDismiss()
    DismissCompanion("CRITTER");
end

--Hook into DismissCompanion() so that we don't auto-recall critters that were dismissed on purpose
RandomCompanion.origDismissCompanion = DismissCompanion;
DismissCompanion = function(...)
    RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter = nil;
    RandomCompanion.StopPetTick()
    debugmsg("Critter manually dismissed. Will not auto-recall.", 4);
    RandomCompanion.origDismissCompanion(...);
end

RandomCompanion.TempDismissCompanion = RandomCompanion.origDismissCompanion;

--Hook into the PetJournal summon/dismiss function so that we don't auto-recall critters that were dismissed on purpose
RandomCompanion.origC_PetJournal_SummonPetByGUID = C_PetJournal.SummonPetByGUID;

C_PetJournal.SummonPetByGUID = function(...)
    local petID = ...;
    
    if petID == C_PetJournal.GetSummonedPetGUID() then
        return DismissCompanion("CRITTER");
    end

    RandomCompanion.SetActiveCompanions(petID);
    return RandomCompanion.origC_PetJournal_SummonPetByGUID(...);
end

function RandomCompanion.Slash(...)
    local parsedcmd = ...;
    local cmd = "";
    local options = "";
    
    debugmsg("cmd: " .. parsedcmd, 1);
    
    if (...) then
        --Parse any macro conditionals such as [modifier:ctrl] or [combat]
        parsedcmd = SecureCmdOptionParse(...);
    end
    
    if parsedcmd == nil then
        parsedcmd = "";
    end
    
    debugmsg("parsed cmd: " .. parsedcmd, 1);
    
    if string.find(parsedcmd, " ") ~= nil then
        cmd = string.sub(parsedcmd, 0, string.find(parsedcmd, " ") - 1);
        options = string.sub(parsedcmd, string.find(parsedcmd, " ") + 1);
    elseif parsedcmd ~= nil then
        cmd = parsedcmd;
    end
    
    if string.lower(cmd) == "autorecall" then
        if (RandomCompanion_Settings.AutoRecall or options == "off") and string.lower(options) ~= "on" then
            RandomCompanion_Settings.AutoRecall = false;
            RandomCompanion_Settings.RandomRecall = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion vanity pet recall disabled");
            if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
                RandomCompanion.StartPetTick()
		    else
		        RandomCompanion.StopPetTick()
            end
        else
            RandomCompanion_Settings.AutoRecall = true;
            RandomCompanion_Settings.RandomRecall = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion vanity pet recall enabled");
            if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
                RandomCompanion.StartPetTick()
		    else
		        RandomCompanion.StopPetTick()
            end
        end
    elseif string.lower(cmd) == "autodismiss" then
        if (RandomCompanion_Settings.AutoDismiss or options == "off") and string.lower(options) ~= "on" then
            RandomCompanion_Settings.AutoDismiss = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion vanity pet auto-dismiss disabled");
            if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
                RandomCompanion.StartPetTick()
		    else
		        RandomCompanion.StopPetTick()
            end
        else
            RandomCompanion_Settings.AutoDismiss = true;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion vanity pet auto-dismiss enabled");
            if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
                RandomCompanion.StartPetTick()
		    else
		        RandomCompanion.StopPetTick()
            end
        end
    elseif string.lower(cmd) == "raiddismiss" then
        if (RandomCompanion_Settings.RaidDismiss or options == "off") and string.lower(options) ~= "on" then
            RandomCompanion_Settings.RaidDismiss = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion vanity pet auto-dismiss disabled");
            if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
                RandomCompanion.StartPetTick()
		    else
		        RandomCompanion.StopPetTick()
            end
        else
            RandomCompanion_Settings.RaidDismiss = true;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion vanity pet auto-dismiss enabled");
            if (RandomCompanion_Settings.AutoRecall or RandomCompanion_Settings.RandomRecall or RandomCompanion_Settings.AutoDismiss or RandomCompanion_Settings.RaidDismiss or RandomCompanion_Settings.RandomChange) and RandomCompanion_Settings[GetRealmName()][UnitName("player")].ActiveCritter then
                RandomCompanion.StartPetTick()
		    else
		        RandomCompanion.StopPetTick()
            end
        end
    elseif string.lower(cmd) == "debug" then    
        if (options ~= "" and tonumber(options) == 0) or (RandomCompanion_Settings.DEBUG and (tonumber(options) == nil or options == "")) then
            if (tonumber(options) ~= nil) then
                RandomCompanion_Settings.DEBUGLevel = options;
            end
            RandomCompanion_Settings.DEBUG = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion debug disabled");
        else
            if (tonumber(options) ~= nil) then
                RandomCompanion_Settings.DEBUGLevel = options;
            end
            if tonumber(RandomCompanion_Settings.DEBUGLevel) < 1 or tonumber(RandomCompanion_Settings.DEBUGLevel) == nil then
                RandomCompanion_Settings.DEBUGLevel = 1;
            end
            RandomCompanion_Settings.DEBUG = true;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion debug enabled - Level " .. RandomCompanion_Settings.DEBUGLevel);
        end
    elseif string.lower(cmd) == "randomchange" then    
        if (options ~= "" and tonumber(options) == 0) or (RandomCompanion_Settings.RandomChange and (tonumber(options) == nil or options == "")) then
            RandomCompanion_Settings.RandomChange = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion random vanity pet change disabled");
        else
            if (tonumber(options) ~= nil) then
                RandomCompanion_Settings.RandomChangeTime = options;
            end
            if tonumber(RandomCompanion_Settings.RandomChangeTime) < 1 or tonumber(RandomCompanion_Settings.RandomChangeTime) == nil then
                RandomCompanion_Settings.RandomChangeTime = 5;
            end
            RandomCompanion_Settings.RandomChange = true;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion random vanity pet change every " .. RandomCompanion_Settings.RandomChangeTime .. " minutes");
        end
    elseif string.lower(cmd) == "mount" then
        RandomCompanion.Mount(options);
    elseif string.lower(cmd) == "pet" then
        RandomCompanion.Pet(options);
    elseif string.lower(cmd) == "petdismiss" or string.lower(cmd) == "dismiss" then
        RandomCompanion.PetDismiss(options);
    elseif string.lower(cmd) == "randomrecall" then
        if (RandomCompanion_Settings.RandomRecall or options == "off") and string.lower(options) ~= "on" then
            RandomCompanion_Settings.AutoRecall = false;
            RandomCompanion_Settings.RandomRecall = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion random vanity pet recall disabled");
        else
            RandomCompanion_Settings.AutoRecall = false;
            RandomCompanion_Settings.RandomRecall = true;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion random vanity pet recall enabled");
        end
    elseif string.lower(cmd) == "cloning" then
        if (RandomCompanion_Settings.Cloning or options == "off") and string.lower(options) ~= "on" then
            RandomCompanion_Settings.Cloning = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion pet and mount cloning disabled");
        else
            RandomCompanion_Settings.Cloning = true;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion pet and mount cloning enabled");
        end
    elseif string.lower(cmd) == "traincrusher" then
        if (RandomCompanion_Settings.Traincrusher or options == "off") and string.lower(options) ~= "on" then
            RandomCompanion_Settings.Traincrusher = false;
            RandomCompanion.EventFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion traincrusher disabled :(");
        else
            if #RandomCompanion.critters.traincrusher then
                RandomCompanion_Settings.Traincrusher = true;
                debugmsg("Parsing combat log events for traincrusher feature", 2);
                RandomCompanion.EventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
                DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion traincrusher enabled!");
            else
                RandomCompanion.EventFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
                DEFAULT_CHAT_FRAME:AddMessage("No traincrushing pets available. :(");
            end
        end
    elseif string.lower(cmd) == "reload" then
        RandomCompanion.Initialize(true);
    elseif string.lower(cmd) == "status" then
        RandomCompanion.Status();
    elseif string.lower(cmd) == "quiet" then
        if (RandomCompanion_Settings.Quiet or options == "off") and string.lower(options) ~= "on" then
            RandomCompanion_Settings.Quiet = false;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion quiet mode disabled");
        else
            RandomCompanion_Settings.Quiet = true;
            DEFAULT_CHAT_FRAME:AddMessage("RandomCompanion quiet mode enabled");
        end
    elseif parsedcmd == ... then --Show help
    
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 RandomCompanion version " .. VERSION);
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 Bind a key from the keybinding menu or...");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc mount\" to choose a random mount");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc mount ground\" to choose a random ground mount in a flyable zone");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc mount passenger\" or \"/rc mount passengerground\" to choose a mount that can carry passengers");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc pet\" to choose a random vanity pet");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc autorecall\" to toggle automatically recalling your vanity pet after resurrecting, changing zones, or taking flight paths");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc randomrecall\" to toggle recalling a random vanity pet after resurrecting, changing zones, or taking flight paths");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc randomchange [number of minutes]\" to toggle randomly changing your vanity pet occasionally. Default is every 15 minutes");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc dismiss\" to dismiss your current vanity pet");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc autodismiss\" to toggle automatically dismissing your vanity pet when you are stealthed and flagged for PVP");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc raidautodismiss\" to toggle automatically dismissing your vanity pet when you are in a raid");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc cloning\" to toggle pet and mount cloning");
        DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00 \"/rc traincrusher\" to toggle Traincrusher!");
    end
    
    if (RandomCompanionOptionsCB_AutoRecall) then
        RandomCompanionOptionsCB_AutoRecall:SetChecked(RandomCompanion_Settings.AutoRecall);
    end
    if (RandomCompanionOptionsCB_AutoDismiss) then
        RandomCompanionOptionsCB_AutoDismiss:SetChecked(RandomCompanion_Settings.AutoDismiss);
    end
    if (RandomCompanionOptionsCB_RaidDismiss) then
        RandomCompanionOptionsCB_RaidDismiss:SetChecked(RandomCompanion_Settings.RaidDismiss);
    end
    if (RandomCompanionOptionsCB_RandomRecall) then
        RandomCompanionOptionsCB_RandomRecall:SetChecked(RandomCompanion_Settings.RandomRecall);
    end
    if (RandomCompanionOptionsCB_Cloning) then
        RandomCompanionOptionsCB_Cloning:SetChecked(RandomCompanion_Settings.Cloning);
    end
    --if (RandomCompanionOptionsCB_RandomChange) then
    --    RandomCompanionOptionsCB_RandomChange:SetChecked(RandomCompanion_Settings.RandomChange);
    --end
end

if not RandomCompanion_Settings then
    RandomCompanion_Settings = {};
end

SlashCmdList["RANDOMMOUNT"] = RandomCompanion.Mount;
SLASH_RANDOMMOUNT1 = "/randommount";
SLASH_RANDOMMOUNT2 = "/rmount";
SLASH_RANDOMMOUNT3 = "/rmt";

SlashCmdList["RANDOMPET"] = RandomCompanion.Pet;
SLASH_RANDOMPET1 = "/randompet";
SLASH_RANDOMPET2 = "/rpet";
SLASH_RANDOMPET3 = "/rpt";

SlashCmdList["PETDISMISS"] = RandomCompanion.PetDismiss;
SLASH_PETDISMISS1 = "/petdismiss";
SLASH_PETDISMISS2 = "/pdismiss";

SlashCmdList["RCSTATUS"] = RandomCompanion.Status;
SLASH_RCSTATUS1 = "/rcstatus";

SlashCmdList["RCDEBUG"] = RandomCompanion.Debug;
SLASH_RCDEBUG1 = "/rcdebug";

SlashCmdList["RCRELOAD"] = RandomCompanion.Initialize;
SLASH_RCRELOAD1 = "/rcreload";

SlashCmdList["RC"] = RandomCompanion.Slash;
SLASH_RC1 = "/rc";
SLASH_RC2 = "/randomcompanion";

RandomCompanion.EventFrame = CreateFrame("Frame", "RandomCompanionFrame");
RandomCompanion.EventFrame:SetScript("OnEvent", RandomCompanion.Event);
RandomCompanion.EventFrame:RegisterEvent("COMPANION_LEARNED");
RandomCompanion.EventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
RandomCompanion.EventFrame:RegisterEvent("MOUNT_JOURNAL_USABILITY_CHANGED");
RandomCompanion.EventFrame:RegisterEvent("PET_JOURNAL_LIST_UPDATE");
RandomCompanion.EventFrame:RegisterEvent("PET_JOURNAL_PET_DELETED");